import logging
import os

from minio import Minio

from udata_hydra import config

log = logging.getLogger("udata-hydra")


class MinIOClient:
    def __init__(self, bucket=config.MINIO_BUCKET):
        self.url = config.MINIO_URL
        self.user = config.MINIO_USER
        self.password = config.MINIO_PWD
        self.bucket = bucket
        self.client = Minio(
            self.url or "test",
            access_key=self.user or "test",
            secret_key=self.password or "test",
            secure=True,
        )
        if self.bucket:
            self.bucket_exists = self.client.bucket_exists(self.bucket)
            if not self.bucket_exists:
                raise ValueError(f"Bucket '{self.bucket}' does not exist.")

    def send_file(
        self,
        file_path,
        delete_source=True,
    ):
        if self.bucket is None:
            raise AttributeError("A bucket has to be specified.")
        is_file = os.path.isfile(os.path.join(file_path))
        if is_file:
            dest_path = f"{config.MINIO_FOLDER}{file_path.split('/')[-1]}"
            self.client.fput_object(
                self.bucket,
                dest_path,
                file_path,
            )
            if delete_source:
                os.remove(file_path)
        else:
            raise Exception(f"file '{file_path}' does not exists")



import gzip
import hashlib
import logging
import tempfile
from typing import IO, Optional

import aiohttp
import magic

from udata_hydra import config

log = logging.getLogger("udata-hydra")


def compute_checksum_from_file(filename: str) -> str:
    """Compute sha1 in blocks"""
    sha1sum = hashlib.sha1()
    with open(filename, "rb") as f:
        block = f.read(2**16)
        while len(block) != 0:
            sha1sum.update(block)
            block = f.read(2**16)
    return sha1sum.hexdigest()


def read_csv_gz(file_path: str) -> IO[bytes]:
    with gzip.open(file_path, "rb") as gz_file:
        with tempfile.NamedTemporaryFile(mode="wb", delete=False) as temp_file:
            temp_file.write(gz_file.read())
    return temp_file


async def download_resource(
    url: str,
    headers: dict,
    max_size_allowed: Optional[int],
) -> IO[bytes]:
    """
    Attempts downloading a resource from a given url.
    Returns the downloaded file object.
    Raises IOError if the resource is too large.
    """
    tmp_file = tempfile.NamedTemporaryFile(
        dir=config.TEMPORARY_DOWNLOAD_FOLDER or None, delete=False
    )

    if max_size_allowed is not None and float(headers.get("content-length", -1)) > max_size_allowed:
        raise IOError("File too large to download")

    chunk_size = 1024
    i = 0
    async with aiohttp.ClientSession(
        headers={"user-agent": config.USER_AGENT}, raise_for_status=True
    ) as session:
        async with session.get(url, allow_redirects=True) as response:
            async for chunk in response.content.iter_chunked(chunk_size):
                if max_size_allowed is None or i * chunk_size < max_size_allowed:
                    tmp_file.write(chunk)
                else:
                    tmp_file.close()
                    log.warning(f"File {url} is too big, skipping")
                    raise IOError("File too large to download")
                i += 1
    tmp_file.close()
    if magic.from_file(tmp_file.name, mime=True) in [
        "application/x-gzip",
        "application/gzip",
    ]:
        tmp_file = read_csv_gz(tmp_file.name)
    return tmp_file



import logging
import os
from pathlib import Path

import toml

log = logging.getLogger("udata-hydra")


class Configurator:
    """Loads a dict of config from TOML file(s) and behaves like an object, ie config.VALUE"""

    configuration: dict = {}

    def __init__(self):
        if not self.configuration:
            self.configure()
            self.load_pyproject_info()

    def configure(self) -> None:
        # load default settings
        configuration: dict = toml.load(Path(__file__).parent / "config_default.toml")

        # override with local settings
        local_settings = os.environ.get("HYDRA_SETTINGS", Path.cwd() / "config.toml")
        if Path(local_settings).exists():
            configuration.update(toml.load(local_settings))

        self.configuration = configuration
        self.check()

    def load_pyproject_info(self) -> None:
        """Get more info about the app from pyproject.toml"""
        try:
            project_info: dict = toml.load("pyproject.toml")["project"]
        except Exception as e:
            log.error(f"Error while getting pyproject.toml info: {str(e)}")
        self.configuration["APP_NAME"] = project_info.get("name")
        self.configuration["APP_VERSION"] = project_info.get("version")

    def override(self, **kwargs) -> None:
        self.configuration.update(kwargs)
        self.check()

    def check(self) -> None:
        """Sanity check on config"""
        assert self.MAX_POOL_SIZE >= self.BATCH_SIZE, "BATCH_SIZE cannot exceed MAX_POOL_SIZE"

    def __getattr__(self, __name):
        return self.configuration.get(__name)

    @property
    def __dict__(self):
        return self.configuration


config = Configurator()



from udata_hydra import context
from udata_hydra.logger import setup_logging

log = setup_logging()


def enqueue(fn, *args, **kwargs):
    """
    Enqueue a task
    Simple helper useful to facilitate mock in testing
    Should be used like this (import matters):
        from udata_hydra import queue
        queue.enqueue(fn, a, b=b, _priority="low")
    """
    priority = kwargs.pop("_priority", "default")
    return context.queue(priority).enqueue(fn, *args, **kwargs)



import json
from typing import Optional

import aiohttp
from aiohttp import web
from marshmallow import ValidationError

from udata_hydra import config, context
from udata_hydra.crawl import check_url
from udata_hydra.db.check import Check
from udata_hydra.db.resource import Resource
from udata_hydra.schemas import CheckSchema
from udata_hydra.utils import get_request_params


async def get_latest_check(request: web.Request) -> web.Response:
    """Get the latest check for a given URL or resource_id"""
    url, resource_id = get_request_params(request, params_names=["url", "resource_id"])
    data: Optional[dict] = await Check.get_latest(url, resource_id)
    if not data:
        raise web.HTTPNotFound()
    if data["deleted"]:
        raise web.HTTPGone()
    return web.json_response(CheckSchema().dump(dict(data)))


async def get_all_checks(request: web.Request) -> web.Response:
    url, resource_id = get_request_params(request, params_names=["url", "resource_id"])
    data: Optional[list] = await Check.get_all(url, resource_id)
    if not data:
        raise web.HTTPNotFound()
    return web.json_response([CheckSchema().dump(dict(r)) for r in data])


async def create_check(request: web.Request) -> web.Response:
    """Create a new check"""

    # Get resource_id from request
    try:
        payload: dict = await request.json()
        resource_id: str = payload["resource_id"]
    except ValidationError as err:
        raise web.HTTPBadRequest(text=json.dumps(err.messages))
    except KeyError as e:
        raise web.HTTPBadRequest(text=f"Missing key: {e}")

    # Get URL from resource_id
    try:
        resource: dict = await Resource.get(resource_id, "url")
        url: str = resource["url"]
    except Exception:
        raise web.HTTPNotFound(text=f"Couldn't find URL for resource {resource_id}")

    context.monitor().set_status(f'Crawling url "{url}"...')

    async with aiohttp.ClientSession(
        timeout=None, headers={"user-agent": config.USER_AGENT}
    ) as session:
        status: str = await check_url(
            url=url, resource_id=resource_id, session=session, worker_priority="high"
        )
        context.monitor().refresh(status)

    check: Optional[dict] = await Check.get_latest(url, resource_id)
    if not check:
        raise web.HTTPBadRequest(text=f"Check not created, status: {status}")

    return web.json_response(CheckSchema().dump(dict(check)))



from .auth import token_auth_middleware
from .csv import detect_tabular_from_headers
from .file import compute_checksum_from_file, download_resource, read_csv_gz
from .http import get_request_params, send
from .queue import enqueue
from .reader import Reader, generate_dialect
from .timer import Timer



from aiohttp import web

from udata_hydra.routes.checks import create_check, get_all_checks, get_latest_check
from udata_hydra.routes.resources import (
    create_resource,
    delete_resource,
    get_resource,
    update_resource,
)
from udata_hydra.routes.status import get_crawler_status, get_health, get_stats, get_worker_status

# routes = web.RouteTableDef()
routes: list = [
    # Routes for checks
    web.get("/api/checks/latest/", get_latest_check),
    web.get("/api/checks/all/", get_all_checks),
    web.post("/api/checks/", create_check),
    # Routes for resources
    web.get("/api/resources/", get_resource),
    web.post("/api/resources/", create_resource),
    web.put("/api/resources/", update_resource),
    web.delete("/api/resources/", delete_resource),
    web.post("/api/resource/created/", create_resource),  # TODO: legacy, to remove
    web.post("/api/resource/updated/", update_resource),  # TODO: legacy, to remove
    web.post("/api/resource/deleted/", delete_resource),  # TODO: legacy, to remove
    # Routes for statuses
    web.get("/api/status/crawler/", get_crawler_status),
    web.get("/api/status/worker/", get_worker_status),
    web.get("/api/stats/", get_stats),
    web.get("/api/health/", get_health),
]



import os

from aiohttp import web

from udata_hydra import context
from udata_hydra.routes import routes
from udata_hydra.utils import token_auth_middleware


async def app_factory() -> web.Application:
    async def app_startup(app):
        app["pool"] = await context.pool()

    async def app_cleanup(app):
        if "pool" in app:
            await app["pool"].close()

    app = web.Application(middlewares=[token_auth_middleware(exclude_methods=("GET",))])
    app.add_routes(routes)
    app.on_startup.append(app_startup)
    app.on_cleanup.append(app_cleanup)
    return app


if __name__ == "__main__":
    web.run_app(app_factory(), path=os.environ.get("HYDRA_APP_SOCKET_PATH"))
